@@ -14,28 +14,22 @@ import traceback
 import numbers
 import string
 import logging
+import dill
+import pathos.pools as pools
 
 from . import toolbox
+from . import VERSION
 from concurrent import futures
 
-VERSION = '0.9.9.6'
-
-cfg = argparse.Namespace()
-cfg.runtime = argparse.Namespace()
-
 
 # handle command line args
 def parse_arguments():
-#    global cfg
-#    cfg.runtime = parser.parse_args(sys.argv[1:], cfg.runtime)
-#    cfg.logformat = "%(message)s"
-#    cfg.logformat = "[%(levelname)s] %(message)s"
     parser = argparse.ArgumentParser(description='A program to run a set of tests for a programming assignment.')
     parser.add_help = True
     parser.add_argument('suite_path', nargs='?', default="./", help='path of test suite to load')
     parser.add_argument('target_path', nargs='?', default="Projects", help='path of target projects (by subdirectory)')
     parser.add_argument('-V', '--version', action='version', version='%(prog)s ' + VERSION)
-    parser.add_argument('-t', '--threads', dest='threads', action='store_true', help='use threads instead of processes')
+    parser.add_argument('-t', '--threaded', dest='threaded', action='store_true', help='use threads instead of processes')
     parser.add_argument('-q', '--quiet', dest='INFO', action='store_false', help='execute in quiet mode (console)')
     parser.add_argument('-w', '--warn', dest='WARN', action='store_true', help='display warning information (console)')
     parser.add_argument('-d', '--debug', dest='DEBUG', action='store_true', help='capture debug information (logfile)')
@@ -43,12 +37,11 @@ def parse_arguments():
     config = parser.parse_args(sys.argv[1:])
     config.logformat = "%(message)s"
     return config
-#    cfg.runtime = parser.parse_args(sys.argv[1:], cfg.runtime)
-#    cfg.logformat = "[%(levelname)s] %(message)s"
 
 
 def build_project(source_root, build_root, build_cfg):
     result_error = None
+    error_output = None
     current_dir = os.getcwd()
 
     # If there is a not a specified build directory, fall back to the source directory instead.
@@ -56,11 +49,12 @@ def build_project(source_root, build_root, build_cfg):
         build_root = source_root
 
     # If the build exists, remove old project folder, recreate it, and switch to it.
-    if os.path.isdir(build_root) and not build_root == source_root:
-        shutil.rmtree(build_root)
-    if not os.path.exists(build_root):
-        os.makedirs(build_root)
-    os.chdir(build_root)
+    if build_root:
+        if os.path.isdir(build_root) and not build_root == source_root:
+            shutil.rmtree(build_root)
+        if not os.path.exists(build_root):
+            os.makedirs(build_root)
+        os.chdir(build_root)
 
     try:
         # Prepare to make substitutions to the prep / build commands if applicable.
@@ -99,72 +93,75 @@ def build_project(source_root, build_root, build_cfg):
                     post_cmd.append(template.substitute(**replacements))
                 subprocess.check_output(post_cmd, stderr=subprocess.STDOUT)
 
-    except (subprocess.CalledProcessError, FileNotFoundError) as error:
+    except subprocess.CalledProcessError as error:
+        result_error = error
+        error_output = error.output
+
+    except FileNotFoundError as error:
         result_error = error
 
     os.chdir(current_dir)
-    return result_error
+    return result_error, error_output
 
 
-def run_suite_tests(framework, subject, proj_settings):
+def run_suite_tests(subject, framework, cfg):
     results = []
     exception_sets = {}
 
-    print("In suite for", subject)
     # Run each project's tests.
-    for project in proj_settings.projects:
-        display_name, identifier, points = project
-        data_set, score, penalty_totals, exception_list = run_project_tests(identifier, framework, subject, proj_settings)
-        exception_sets[project] = exception_list
+    for test_set in cfg.sets:
+        data_set, score, penalty_totals, exception_list = run_test_set(test_set, subject, framework, cfg)
+        exception_sets[test_set.name] = exception_list
 
         # If the project didn't compile, just add a single line indicating that.
         if isinstance(data_set, str):
-            logging.error("Error: %s" % data_set)
+            exception_sets[test_set.name].append(data_set)
             data_set = [[], ["Grade: 0 (Does not compile / run)"]]
             score = 0
 
         else:
             # Add info on the score and penalty values for the project.
-            data_set += [ [], ["Test Cases: %.2f (%.2f%%)" % (score * points, score * 100) ] ]
+            data_set += [ [], ["Test Cases: %.2f (%.2f%%)" % (score * test_set.max_score, score * 100) ] ]
             overall_penalty = 0
 
-            for penalty_num in range(0, len(proj_settings.test_case_penalties)):
-                penalty_name, magnitude = proj_settings.test_case_penalties[penalty_num]
+            for penalty_num, case_penalty in enumerate(test_set.case_penalties):
+                penalty_name, magnitude, _ = case_penalty
                 overall_penalty += penalty_totals[penalty_num]
                 data_set += [ [ "%s Penalty (overall): %.2f%%" % (penalty_name, penalty_totals[penalty_num] * 100) ] ]
 
-            for penalty_num in range(0, len(proj_settings.project_penalties)):
-                penalty_name, magnitude = proj_settings.project_penalties[penalty_num]
-                penalty_ind = penalty_num + len(proj_settings.test_case_penalties)
+            for penalty_num, set_penalty in enumerate(test_set.set_penalties):
+                penalty_name, magnitude, _ = test_set.set_penalties[penalty_num]
+                penalty_ind = penalty_num + len(test_set.case_penalties)
                 overall_penalty += penalty_totals[penalty_ind]
                 data_set += [ [ "%s Penalty (overall): %.2f%%" % (penalty_name, penalty_totals[penalty_ind] * 100) ] ]
 
             # Apply the penalties and scale to the number of points
-            score = (score - min(proj_settings.max_penalty, overall_penalty)) * points
+            score = (score - min(test_set.max_penalty, overall_penalty)) * test_set.max_score
 
         # Add to the results list.
-        data_set = [ ["Project %s" % display_name] ] + data_set
-        results.append((display_name, score, data_set))
+        data_set = [["Test-Set %s" % test_set.name]] + data_set
+        results.append((test_set.name, score, data_set))
 
     return results, exception_sets
 
 
-def run_project_tests(name, framework, subject, proj_settings):
+def run_test_set(test_set, subject, framework, cfg):
+    # Prepare data structures and initialize the test set.
     exception_list = []
-    context = proj_settings.initialize_project(name, framework, subject, proj_settings)
-    penalty_totals = [0] * (len(proj_settings.test_case_penalties) + len(proj_settings.project_penalties))
-    num_of_tests = proj_settings.get_number_of_tests(context)
+    set_context = cfg.initialize_test_set(test_set, subject, framework)
+    num_of_tests = test_set.get_num_tests(set_context, subject, framework, cfg)
+    penalty_totals = [0] * (len(test_set.case_penalties) + len(test_set.set_penalties))
 
     if not isinstance(num_of_tests, int) or num_of_tests == 0:
-        proj_settings.shutdown_project(context)
-        return "get_number_of_tests() returned [%s]" % num_of_tests, None, None
+        cfg.shutdown_test_set(set_context)
+        return "get_number_of_tests() returned [%s]" % num_of_tests, None, None, exception_list
 
     # Add the initial notes at the top (might adjust this later for 'pure' CSV output)
     data_set = [ [ "Number of tests: %d." % num_of_tests ], [] ]
 
     # Prepare the header.
     header = [ 'Test No.', 'Score', 'Message', 'Desc.' ]
-    header.extend(["%s-Pen" % name for name, pct in proj_settings.test_case_penalties])
+    header.extend(["%s-Pen" % penalty[0] for penalty in test_set.case_penalties])
     data_set.append(header)
 
     score = 0
@@ -172,7 +169,7 @@ def run_project_tests(name, framework, subject, proj_settings):
         # Set up the row for this test and run it.
         row = [ '%d' % test_num ]
         try:
-            case_result = proj_settings.run_case_test(test_num, context)
+            case_result = test_set.run_case_test(test_num, set_context, subject, framework, cfg)
         except Exception as e:
             stack_trace = traceback.format_exc()
             exception_list.append("Test %d, %s: %s\n%s" % (test_num, type(e).__name__, e, stack_trace))
@@ -191,86 +188,59 @@ def run_project_tests(name, framework, subject, proj_settings):
         # Add score, run message, and description as applicable
         row.append('%.2f%%' % (case_score * 100))
         row.append(message if message else '')
-        row.append(proj_settings.get_test_description(test_num, context) if round(case_score, 10) < 1 else '')
+        row.append(test_set.get_test_desc(test_num, set_context, subject, framework, cfg) if round(case_score, 10) < 1 else '')
 
         if case_score == 0:
             data_set.append(row)
             continue
 
-        for penalty_num in range(0, len(proj_settings.test_case_penalties)):
-            pen_result = proj_settings.test_case_penalties[penalty_num]
-            penalty = proj_settings.run_case_penalty(penalty_num, test_num, context)
-
-            # If we successfuly completed the run, this should be a tuple; otherwise, a message.
-            if isinstance(pen_result, tuple):
-                penalty_name, magnitude = pen_result
-                message = None
-            else:
-                penalty_name = None
-                maginitude = 0
-                message = str(pen_result)
-
+        # Go through each penalty and run it (if valid).
+        for penalty_num, case_penalty in enumerate(test_set.case_penalties):
+            penalty_name, magnitude, pen_function = case_penalty
+            penalty = pen_function(penalty_num, test_num, set_context, subject, framework, cfg)
             penalty_totals[penalty_num] += penalty * magnitude * case_score / num_of_tests
             row.append('%.2f%%' % (penalty * 100))
 
         # Add this test data to the data set.
         data_set.append(row)
 
-    for penalty_num in range(0, len(proj_settings.project_penalties)):
-        penalty_name, magnitude = proj_settings.project_penalties[penalty_num]
-        penalty = proj_settings.run_project_penalty(penalty_num, context) * score * magnitude
-        penalty_totals[penalty_num + len(proj_settings.test_case_penalties)] = penalty
+    for penalty_num, set_penalty in enumerate(test_set.set_penalties):
+        penalty_name, magnitude, pen_function = set_penalty
+        penalty = pen_function(penalty_num, set_context, subject, framework, cfg) * score * magnitude
+        penalty_totals[penalty_num + len(test_set.case_penalties)] = penalty
 
     # Return the data set, score (proportion), and penalty totals
-    proj_settings.shutdown_project(context)
+    cfg.shutdown_test_set(set_context)
     return data_set, score / num_of_tests, penalty_totals, exception_list
 
 
-def process_configuration(config):
-    if not config:
-        return None
-
-    # Fill default values
-    if not hasattr(config.build, 'prep_cmd'):
-        config.build.prep_cmd = []
-    if not hasattr(config.build, 'compile_cmd'):
-        config.build.compile_cmd = []
-
-    # Make configuration paths absolute
-    config.general.result_path = os.path.abspath(config.general.result_path) if config.general.result_path else None
-
-    config.build.base = os.path.abspath(config.build.base) if config.build.base else None
-    config.build.destination = os.path.abspath(config.build.destination) if config.build.destination else None
-    config.build.resources = os.path.abspath(config.build.resources) if config.build.resources else None
-
-    config.build.subject_src = os.path.abspath(config.build.subject_src) if config.build.subject_src else None
-    config.build.subject_bin = os.path.abspath(config.build.subject_bin) if config.build.subject_bin else None
-    config.build.framework_src = os.path.abspath(config.build.framework_src) if config.build.framework_src else None
-    config.build.framework_bin = os.path.abspath(config.build.framework_bin) if config.build.framework_bin else None
-
-    return config
-
-# For each submission, copy the base files, then the submission, into the destination folder.
-def prepare_and_init_framework(build_cfg, proj_cfg):
-    # Build the environment components (only need to do this once.)
-    if build_cfg.framework_src and build_cfg.framework_bin:
-        if build_cfg.prep_cmd or build_cfg.compile_cmd:
+# Build the environment components (only need to do this once.)
+def prepare_and_init_framework(cfg):
+    if cfg.build.framework_src and cfg.build.framework_bin:
+        if cfg.build.prep_cmd or cfg.build.compile_cmd:
             logging.info("Prepping / building framework environment... ")
-            result_error = build_project(build_cfg.framework_src, build_cfg.framework_bin, build_cfg)
+            result_error, error_output = build_project(cfg.build.framework_src, cfg.build.framework_bin, cfg.build)
             if result_error:
-                logging.info("%s: %s\n" % (type(result_error).__name__, result_error))
+                error_text = "%s: %s\n" % (type(result_error).__name__, result_error)
+                logging.error(error_output if error_output else error_text)
+                logging.info(error_text)
                 return
 
         logging.info("initializing framework... ")
-        framework_data = proj_cfg.initialize_framework(cfg)
+        framework_data = cfg.initialize_framework()
         logging.info("done.\n")
     else:
         framework_data = None
     return framework_data
 
 # For each submission, copy the base files, then the submission, into the destination folder.
-def prepare_and_test_submission(framework_context, submission):
-    global cfg
+def prepare_and_test_submission(submission, framework_context, cfg):
+    # Because this might be in a new process, we wil need to reset the console logger when prep the project.
+    console_logger = toolbox.SelectiveStreamHandler(INFO=cfg.runtime.INFO, WARNING=cfg.runtime.WARN, CRITICAL=True)
+    logging.basicConfig(format=cfg.runtime.logformat, level=logging.DEBUG, handlers=[console_logger])
+    root_logger = logging.getLogger('')
+    console_logger.terminator = ""
+
     if not os.path.isdir(submission):
         return None
 
@@ -284,48 +254,60 @@ def prepare_and_test_submission(framework_context, submission):
     shutil.copytree(submission, cfg.build.destination, dirs_exist_ok=True)
 
     # Build the project.
+    setup_exceptions = []
     logging.info("Prepping / building project(s) for " + submission + "... ")
-    error = build_project(cfg.build.subject_src, cfg.build.subject_bin, cfg.build)
+    error, output = build_project(cfg.build.subject_src, cfg.build.subject_bin, cfg.build)
+
     if error:
         logging.info("error building (see logs)... ")
-        logging.error("Error prepping/building %s - %s: %s" % (submission, type(error).__name__, error))
+        error_text = "While prepping/building, %s: %s\n" % (type(error).__name__, error)
+        error_text += output if output else ""
+        setup_exceptions.append(error_text)
 
     logging.info("Initializing... ")
     subject_context = None
     starting_dir = os.getcwd()
 
     try:
-        subject_context = cfg.project.initialize_subject(cfg)
+        subject_context = cfg.initialize_subject(submission, framework_context)
     except Exception as error:
-        logging.error("Error initializing subject %s - %s: %s" % (submission, type(error).__name__, error))
         logging.info("error initializing (see logs)... ")
+        setup_exceptions.append("Error initializing subject %s - %s: %s" % (submission, type(error).__name__, error))
 
     os.chdir(starting_dir)
     logging.info("done.\n")
 
     starting_dir = os.getcwd()
-    results, exception_sets = run_suite_tests(framework_context, subject_context, cfg.project)
-    cfg.project.shutdown_subject(subject_context)
+    results, exception_sets = run_suite_tests(subject_context, framework_context, cfg)
+    cfg.shutdown_subject(subject_context)
     os.chdir(starting_dir)
+    if len(setup_exceptions) > 0:
+        exception_sets["Setup"] = setup_exceptions
+
     return results, exception_sets
 
 
 def main():
-    global cfg # TODO: Eventually take this out of global scope for simplicity
-    cfg.runtime = parse_arguments()
+    dill.settings['recurse']=True
+    runtime = parse_arguments()
 
     # Save the current folder and move to the test suite location.
     starting_dir = os.getcwd()
-    os.chdir(cfg.runtime.suite_path)
+    os.chdir(runtime.suite_path)
 
     # Load the config file for this project.
-    if not os.path.isfile("config.py") or not (config := process_configuration(toolbox.load_module("config.py"))):
-        sys.stderr.write("Error: no configuration file. Exiting...\n")
+    config_path = os.path.join("Settings","config.py")
+
+    if not os.path.isdir("Settings"):
+        sys.stderr.write("Error: no settings folder. Exiting...\n")
         return
 
-    cfg.project = config.project
-    cfg.build = config.build
-    cfg.general = config.general
+    elif not os.path.isfile(config_path) or not (config := toolbox.load_module(config_path)):
+        sys.stderr.write("Error: could not load configuration file. Exiting...\n")
+        return
+
+    cfg = config.get_suite_config(runtime)
+    cfg.make_paths_absolute()
 
     # Prepare result paths.
     if not os.path.isdir(cfg.general.result_path):
@@ -351,13 +333,13 @@ def main():
     # Initialize the framework / get any important info
 #    with futures.ProcessPoolExecutor() as executor:
 #        try:
-#            future = executor.submit(prepare_and_init_framework, cfg.build, cfg.project)
+#            future = executor.submit(prepare_and_init_framework, cfg)
 #            framework_context = future.result()
             # This is a fatal error; if we can't initalize the framework, we should stop here.
 #        except Exception as e:
 #            sys.stderr.write("Error initializing framework - %s: %s. Exiting.\n" % (type(e).__name__, e))
 #            exit()
-    framework_context = prepare_and_init_framework(cfg.build, cfg.project)
+    framework_context = prepare_and_init_framework(cfg)
 
     # Close general log file and move on to student-specific logs.
     root_logger.removeHandler(file_logger)
@@ -376,11 +358,14 @@ def main():
         file_logger = toolbox.SelectiveFileHandler(logfile, mode="w", DEBUG=cfg.runtime.DEBUG, ERROR=True)
         root_logger.addHandler(file_logger)
 
-        exec_class = futures.ThreadPoolExecutor if cfg.runtime.threads else futures.ProcessPoolExecutor
+        exec_class = pools.ThreadPool if cfg.runtime.threaded else pools.ProcessPool
+#        exec_class = futures.ThreadPoolExecutor if cfg.runtime.threaded else futures.ProcessPoolExecutor
         with exec_class() as executor:
             try:
-                future = executor.submit(prepare_and_test_submission, framework_context, submission)
-                suite_results, exception_sets = future.result()
+#                future = executor.submit(prepare_and_test_submission, submission, framework_context, cfg)
+#                suite_results, exception_sets = future.result()
+                future = executor.apipe(prepare_and_test_submission, submission, framework_context, cfg)
+                suite_results, exception_sets = future.get()
                 # If there were exceptions in the tests, we should log them.
                 for project, exception_list in exception_sets.items():
                    if len(exception_list) > 0:
@@ -400,10 +385,10 @@ def main():
 
         if suite_results:
             for name, result, data_set in suite_results:
-                file_data.extend(data_set + [ [ "%s: %.3f" % (name, result) ] ])
+                file_data.extend([[]] + data_set + [["Set Total: %.3f" % result]])
                 grand_total += result
 
-        file_data.append([ "Overall score: %.2f" % grand_total ])
+        file_data.extend([[], ["Overall Score: %.2f" % grand_total]])
         toolbox.save_csv(os.path.join(output_dir, cfg.general.result_file), file_data)
 
         # Add data to summary file for this submission.
@@ -417,8 +402,8 @@ def main():
         file_logger.close()
         time.sleep(2)
 
-    cfg.project.shutdown_framework(framework_context)
-    print("Framework shutdown")
+    cfg.shutdown_framework(framework_context)
+    logging.info("Framework shutdown")
     # Return to where we started at.
     os.chdir(starting_dir)
 
