@@ -128,19 +128,25 @@ def write_to_file(lines, filename):
     start_dir = os.getcwd()
     os.chdir(working_dir)
 
+def load_relative_module(target, neighbor, package_name=None):
+    if not os.path.isabs(target):
+        target = os.path.join(path.abspath(os.path.dirname(neighbor)), target)
+    return load_module(target, package_name)
+
+
 def load_module(filename, package_name=None):
     start_dir = os.getcwd()
     mod_dir, mod_file = path.split(filename)
     sys.path.append('./')
-
-    if mod_dir:
-        os.chdir(mod_dir)
-
-    module_name = (package_name if package_name else '') + path.splitext(mod_file)[0]
-    spec = importlib.util.spec_from_file_location(module_name, mod_file)
-    module = importlib.util.module_from_spec(spec)
+    module = None
 
     try:
+        if mod_dir:
+            os.chdir(mod_dir)
+
+        module_name = (package_name if package_name else '') + path.splitext(mod_file)[0]
+        spec = importlib.util.spec_from_file_location(module_name, mod_file)
+        module = importlib.util.module_from_spec(spec)
         spec.loader.exec_module(module)
     except FileNotFoundError as e:
         print(type(e).__name__ + ": " + str(e))
@@ -198,7 +204,7 @@ def load_library(directory, name):
 
 def unload_library(library):
     if "FreeLibrary" in dir(_ctypes):
-        _ctypes.FreeLibrary(library)
+        _ctypes.FreeLibrary(library._handle)
     else:
         _ctypes.dlclose(library._handle)
 
@@ -492,6 +498,10 @@ def get_value_or_error_msg(target_function, params):
             return e.__class__.__name__
 
 
+def get_type_and_value_string(target):
+    return type(target).__name__ + ": " + str(target)
+
+
 def ansi_to_text(text, lines=30, columns=80):
 #, convert_glyphs=True):
     screen = pyte.Screen(columns, lines)
